open F

type type_error =
  | TypeMismatch of debruijn_type * debruijn_type
  | ArityMismatch of arity_requirement * debruijn_type
  | NotAnArrow of debruijn_type
  | NotAProduct of debruijn_type
  | NotAForall of debruijn_type
  | UnboundTermVariable of tevar
  | TwoOccurencesOfSameVariable of string

and arity_requirement = Index of int | Total of int

exception TypeError of type_error

(* -------------------------------------------------------------------------- *)

(* A type environment maps term variables to types in de Bruijn's representation. *)

(* Our term variables are in nominal representation, which allows us to represent
   the environment as a binary search tree. One difficulty is that when we enter a
   new type variable binder [TyAbs], we must (conceptually) shift every type in
   the range of the environment up by one. When term variables are in de Bruijn
   notation and the environment is a list, this is easily done by inserting a mark
   in front of the environment. Here, we must be more clever. We maintain an [N2DB]
   environment that tells us, for every *term* variable, how long ago it was bound,
   that is, how many *type* variables were introduced since then. Hence, this tells
   us by how much we must shift its type. It's really a bit too clever, but let's
   do this for fun. *)

module TermVar =
  String

module TermVarMap =
  Map.Make(TermVar)

module N2DB =
  DeBruijn.Nominal2deBruijn(TermVar)

type env = {
  (* A mapping of term variables to types. *)
  types: debruijn_type TermVarMap.t;
  (* A translation environment of TERM variables to TYPE indices. *)
  names: N2DB.env;
}

type _datatype_env = (unit, debruijn_type) Datatype.Env.t

let empty =
  { types = TermVarMap.empty; names = N2DB.empty }

let lookup { types; names } x =
  try
    (* Obtain the type associated with [x]. *)
    let ty = TermVarMap.find x types in
    (* Find how long ago [x] was bound. *)
    let w = N2DB.lookup names x in
    (* Shift the type [ty] by this amount, so that it makes sense in the
       current scope. *)
    Type.lift w 0 ty
  with Not_found ->
    (* must have been raised by [TermVarMap.find] *)
    raise (TypeError (UnboundTermVariable x))

let extend_with_tevar { types; names } x ty =
  (* Map the name [x] to [ty], and record when it was bound, without
     incrementing the time. *)
  { types = TermVarMap.add x ty types;
    names = N2DB.slide names x }

let extend_with_tyvar { types; names } =
  (* Increment the time. *)
  { types; names = N2DB.bump names }

let singleton x ty =
  extend_with_tevar empty x ty

let concat env1 env2 =
  let combine _ _ ty2 = Some ty2 in
  { types = TermVarMap.union combine env1.types env2.types ;
    names = N2DB.concat env1.names env2.names }

let concat_disjoint env1 env2 =
  let exception Conflict of TermVarMap.key in
  try
    let combine x _ _ =
      raise (Conflict x) in
    let types = TermVarMap.union combine env1.types env2.types
    and names = N2DB.concat env1.names env2.names in
    Ok { types; names; }
  with Conflict x -> Error x

(* -------------------------------------------------------------------------- *)

(* Destructors. *)

let unfold ty =
  match ty with
  | TyMu ((), body) ->
      Type.subst ty 0 body
  | _ ->
      assert false

let rec as_arrow ty =
  match ty with
  | TyArrow (ty1, ty2) ->
      ty1, ty2
  | TyMu _ ->
      as_arrow (unfold ty)
  | _ ->
      raise (TypeError (NotAnArrow ty))

let rec as_product ty =
  match ty with
  | TyProduct tys ->
      tys
  | TyMu _ ->
      as_product (unfold ty)
  | _ ->
      raise (TypeError (NotAProduct ty))

let rec as_forall ty =
  match ty with
  | TyForall ((), ty) ->
      ty
  | TyMu _ ->
      as_forall (unfold ty)
  | _ ->
      raise (TypeError (NotAForall ty))

(* -------------------------------------------------------------------------- *)

(* An equality test. *)

(* In the absence of recursive types, syntactic equality would work. However,
   in their presence, one must unfold recursive types during the equality
   check. *)

(* Note that comparing two [TyMu] types just by comparing their bodies is
   correct, but incomplete. We do not do it. Naively unfolding every [TyMu]
   type is complete, but leads to potential non-termination. We do that, with
   a budget limit, which makes the algorithm unsound (two types may be
   considered equal when they are different). I should really implement
   my own algorithm (Gauthier & Pottier, ICFP 2004) but that will be for
   some other time. *)

let rec equal budget ty1 ty2 =
  match ty1, ty2 with
  | (TyMu _ as ty1), ty2
  | ty2, (TyMu _ as ty1) ->
      budget = 0 || equal (budget - 1) (unfold ty1) ty2
  | TyVar x1, TyVar x2 ->
      x1 = x2
  | TyVar _, _ ->
      false
  | TyArrow (ty1a, ty1b), TyArrow (ty2a, ty2b) ->
      equal budget ty1a ty2a && equal budget ty1b ty2b
  | TyArrow _, _ ->
      false
  | TyProduct tys1, TyProduct tys2
  | TySum tys1, TySum tys2 ->
      equal_list budget tys1 tys2
  | TyProduct _, _
  | TySum _, _ ->
      false
  | TyForall ((), ty1), TyForall ((), ty2) ->
      equal budget ty1 ty2
  | TyForall _, _ ->
      false
  | TyConstr (Datatype.Type tyconstr1, tys1),
    TyConstr (Datatype.Type tyconstr2, tys2) ->
      String.equal tyconstr1 tyconstr2 &&
      equal_list budget tys1 tys2
  | TyConstr _, _ ->
      false

and equal_list budget tys1 tys2 =
  try
    List.for_all2 (equal budget) tys1 tys2
  with Invalid_argument _ -> false


let budget =
  4

let (--) ty1 ty2 =
  if not (equal budget ty1 ty2) then
    raise (TypeError (TypeMismatch (ty1, ty2)))

(* -------------------------------------------------------------------------- *)

(* The type-checker. *)

let nth_type ty tys i =
  if i < 0 || i >= List.length tys then
    raise (TypeError (ArityMismatch (Index i, ty)))
  else
    List.nth tys i

let rec typeof datatype_env env (t : debruijn_term) : debruijn_type =
  let typeof = typeof datatype_env in
  match t with
  | Var x ->
      lookup env x
  | Abs (x, ty1, t) ->
      let ty2 = typeof (extend_with_tevar env x ty1) t in
      TyArrow (ty1, ty2)
  | App (t, u) ->
      let ty1, ty2 = as_arrow (typeof env t) in
      typeof env u -- ty1;
      ty2
  | Let (x, t, u) ->
      let env = extend_with_tevar env x (typeof env t) in
      typeof env u
  | TyAbs ((), t) ->
      TyForall ((), typeof (extend_with_tyvar env) t)
  | TyApp (t, ty2) ->
      Type.subst ty2 0 (as_forall (typeof env t))
  | Tuple ts ->
      let tys = List.map (typeof env) ts in
      TyProduct tys
  | Proj (i, t) ->
      let ty = typeof env t in
      let tys = as_product ty in
      nth_type ty tys i
  | Inj (tys, i, t) ->
      let ty = typeof env t in
      ty -- nth_type (TySum tys) tys i;
      TySum tys
  | Variant (lbl, dty, t) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      let ty = typeof env t in
      ty -- arg_type;
      TyConstr dty
  | LetProd (xs, t, u) ->
      let ty = typeof env t in
      let tys = as_product ty in
      if List.length xs <> List.length tys then
        raise (TypeError (ArityMismatch (Total (List.length xs), ty)));
      let env = List.fold_left2 extend_with_tevar env xs tys in
      typeof env u
  | Match (ty, t, brs) ->
      let scrutinee_ty = typeof env t in
      let tys = List.map (typeof_branch datatype_env env scrutinee_ty) brs in
      List.iter ((--) ty) tys;
      ty

and typeof_variant datatype_env lbl (tid, tyargs) =
  let Datatype.{ data_kind ; type_params ; labels_descr; _ } =
    Datatype.Env.find_decl datatype_env tid in
  let Datatype.{ arg_type; _ } =
    List.find (fun l -> l.Datatype.label_name = lbl) labels_descr in
  begin
    match data_kind with
    | Datatype.Variant ->
        ()
    | Datatype.Record ->
        raise Datatype.Env.UnexpectedRecord
  end;
  List.fold_right2 (fun ty () arg_type ->
    Type.subst ty 0 arg_type
  ) tyargs type_params arg_type

and typeof_branch datatype_env env scrutinee_ty (pat, rhs) =
  let new_bindings = binding_pattern datatype_env scrutinee_ty pat in
  let new_env = concat env new_bindings in
  typeof datatype_env new_env rhs

and binding_pattern datatype_env scrutinee_ty pat =
  let binding_pattern = binding_pattern datatype_env in
  match pat with
  | PVar x ->
      singleton x scrutinee_ty
  | PWildcard ->
      empty
  | PInj (tys, i, pat) ->
      scrutinee_ty -- TySum tys;
      let ty' = nth_type (TySum tys) tys i in
      binding_pattern ty' pat
  | PTuple ps ->
      let tys = as_product scrutinee_ty in
      let envs = List.map2 binding_pattern tys ps in
      let f env1 env2 =
        match concat_disjoint env1 env2 with
        | Ok env -> env
        | Error x -> raise (TypeError (TwoOccurencesOfSameVariable x))
      in
      List.fold_left f empty envs
  | PVariant (lbl, dty, pat) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      scrutinee_ty -- TyConstr dty;
      binding_pattern arg_type pat

let typeof datatype_env t =
  typeof datatype_env empty t
