%{
  open ML
%}

%token <string> LIDENT
%token <string> UIDENT
%token <string> QIDENT

%token FUN
%token ARROW "->"
%token LET IN
%token EQ "="

%token LPAR "("
%token RPAR ")"

%token TYPE
%token OF
%token LBRACE "{"
%token RBRACE "}"
%token STAR "*"

%token COMMA
%token WILDCARD
%token PIPE
%token MATCH WITH END

%token COLON ":"
%token PERIOD "."
%token SOME

%token EOF

%start <(ML.datatype_env * (string * ML.term) list)> prog

%%

let prog :=
  | datatype_env = type_decls ; xts = nonempty_list (toplevel_term_decl) ; EOF ;
                                          { (datatype_env, xts) }

(***************** TERMS ***************)
let toplevel_term_decl :=
  | LET ; x = tevar_ ; "=" ; t = term ;   { (x, t) }

let term :=
  | ~ = term_abs ; <>

let term_abs :=
  | FUN ; x = tevar_ ; "->" ; t = term_abs ;
                                          { Abs (x, t) }

  | (x, t1, t2) = letin(tevar) ;          { Let (x, t1, t2) }

  | (xs, t1, t2) = letin(tuple(tevar)) ;  { LetProd (xs, t1, t2) }

  | ~ = term_app ; <>

let term_app :=
  | t1 = term_app ; t2 = term_atom ;
    {
      match t1 with
      | Variant (l, Tuple []) ->
	 Variant (l, t2)
      | _ ->
	 App (t1, t2)
    }

  | ~ = term_atom ; <>

let term_atom :=
  | x = tevar ;                           { Var x }

  | l = UIDENT ;                          { Variant (Datatype.Label l, Tuple []) }

  | ts = tuple (term) ;                   { Tuple ts }

  | MATCH ; t = term ; WITH ;
    brs = branches ;
    END ;                                 { Match (t, brs) }

  | "(" ; t = term ; ":" ; tyannot = type_annotation ; ")" ;
                                          { Annot (t, tyannot) }

  | "(" ; ~ = term ; ")" ; <>

let branches :=
  | ~ = pipe_separated_list (branch) ; <>

let branch :=
  | pat = pattern ; "->" ; t = term ;     { (pat, t) }

let pattern :=
  | l = UIDENT ; pat = pattern_atom ;     { PVariant (Datatype.Label l, pat) }

  | ~ = pattern_atom ; <>

let pattern_atom :=
  | x = tevar ;                           { PVar x }

  | WILDCARD ;                            { PWildcard }

  | ps = tuple (pattern) ;                { PTuple ps }

  | "(" ; p = pattern ; ":" ; tyannot = type_annotation ; ")" ;
                                          { PAnnot (p, tyannot) }

let tevar :=
  | ~ = LIDENT ; <>

let tevar_ :=
  | ~ = LIDENT ; <>
  | WILDCARD ;   { "_" }

(*************** TYPES ***************)

let typ :=
  | ~ = type_arrow ; <>

let type_arrow :=
  | ty1 = type_tyconstr ; "->" ; ty2 = type_arrow ;
                                          { TyArrow (ty1, ty2) }

  | ~ = type_tyconstr ; <>


let type_tyconstr :=
  | ~ = tyname ; typarams = list (type_atom) ;
                                          { TyConstr (tyname, typarams)}

  | ~ = type_atom ; <>

let type_atom :=
  | x = tyvar ;                           { TyVar x }

  | "{" ; tys = separated_list ("*", typ) ; "}" ;
                                          { TyProduct tys }

  | "(" ; ~ = typ ; ")" ;                 <>


let type_annotation :=
  | SOME ; xs = list(tyvar) ; "." ; ty = typ ;
                                          { (xs, ty) }

let tyconstr_decl :=
  | l = UIDENT ; OF ; arg_type = typ ;
                                          { (Datatype.Label l, arg_type) }

let type_decls :=
  | decls = list (type_decl) ;
    { List.fold_left Datatype.Env.add_decl Datatype.Env.empty decls }

let type_decl :=
  | TYPE ; ~ = tyname ; type_params = list (tyvar) ; "=" ;
    arg_types = pipe_separated_list (tyconstr_decl) ;
    {
      let name = tyname in
      let data_kind = Datatype.Variant in
      let labels_descr = List.map (fun (label_name, arg_type) ->
          Datatype.{ label_name ; type_name = name ; arg_type } )
        arg_types
      in
      Datatype.{ name ; type_params ; data_kind ; labels_descr }
    }

let tyname :=
  | lid = LIDENT ; { Datatype.Type lid }

let tyvar :=
  | ~ = QIDENT ; <>

let tuple (X) :=
  | "(" ; ")" ;                           { [] }
  | "(" ; x = X ; COMMA ; xs = inside_tuple (X) ; ")" ;
                                          { x :: xs }

let inside_tuple (X) :=
  |                                       { [] }
  | x = X ;                               { [x] }
  | x = X ; COMMA ; xs = inside_tuple(X); { x :: xs }

let letin (X) :=
  | LET ; x = X ; EQ ;
      t1 = term ; IN ;
      t2 = term_abs ;                     { (x, t1, t2) }

let pipe_separated_list (X) :=
  | option (PIPE) ; ~ = separated_list (PIPE, X) ; <>
